from telluric.georaster import GeoRaster2
from telluric.vectors import GeoVector
from telluric.features import GeoFeature
from telluric.collections import FileCollection, FeatureCollection

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
